from .collection_resolver import CollectionResolver
from ...entities.wrappers import KCScalar
from ...functional.util import trace_to_full_accessor

class MissingValueResolver(CollectionResolver):
    def preorder_resolve(self, node, root_node, result, trace, root_resolver):
        if isinstance(node, KCScalar) and isinstance(node.value, str) and node.value == "???":
            from ...errors import missing_value_error
            raise missing_value_error(trace_to_full_accessor(trace))



import yaml
from ..entities.grammar_tree_nodes import RootNode, InterpolatedNode, FixedNode
from ..entities.wrappers import KCScalar
from ..functional.load import from_string

def parse_grammar(value):
    root_node = RootNode()
    _parse(value, root_node)
    return root_node


def _parse(value, parent_node):
    if isinstance(value, KCScalar):
        value = value.value
    if not isinstance(value, str):
        from ..errors import unexpected_type_error
        raise unexpected_type_error(str, value)

    while True:
        if len(value) == 0:
            return
        if "${" not in value:
            parent_node.children.append(FixedNode(value))
            return

        brace_start = value.index("${")
        if brace_start != 0:
            parent_node.children.append(FixedNode(value[:brace_start]))
        value = value[brace_start + 2:]
        brace_end = _find_brace_end(value)

        value_in_braces = value[:brace_end]
        colon_idx = _find_colon(value_in_braces)
        if colon_idx == -1:
            if len(value_in_braces) == 0:
                from ..errors import empty_resolver_value_error
                raise empty_resolver_value_error(value_in_braces)
            node = InterpolatedNode(None)
        elif colon_idx == 0:
            from ..errors import empty_resolver_key_error
            raise empty_resolver_key_error(value_in_braces)
        else:
            if len(value_in_braces) == colon_idx + 1:
                from ..errors import empty_resolver_value_error
                raise empty_resolver_value_error(value_in_braces)
            node = InterpolatedNode(value_in_braces[:colon_idx])
        parent_node.children.append(node)

        # parse recursively
        _parse(value[colon_idx + 1:brace_end], node)
        value = value[brace_end + 1:]


def _find_brace_end(value):
    nesting_level = 1
    for i in range(len(value)):
        if value[i] == "}":
            nesting_level -= 1
            if nesting_level == 0:
                return i
        if value[i] == "$" and i != len(value) - 2 and value[i + 1] == "{":
            nesting_level += 1
    from ..errors import missing_closing_brace_error
    raise missing_closing_brace_error(value)

def _find_colon(value):
    if "${" in value:
        value = value[:value.index("${")]
    if ":" not in value:
        return -1
    return value.index(":")

def parse_resolver_args_and_value(args_and_value, n_args):
    args = []
    remaining_str = args_and_value
    for i in range(n_args):
        if not ":" in remaining_str:
            from ..errors import missing_parameter_error
            raise missing_parameter_error(args_and_value, n_args)
        colon_idx = remaining_str.index(":")
        arg_str = remaining_str[:colon_idx]
        if len(arg_str) == 0:
            from ..errors import empty_parameter_error
            raise empty_parameter_error(args_and_value)

        # parse into primitives (e.g. parse '5' to 5)
        # also removes leading/trailing whitespaces for strings
        parsed_arg = yaml.safe_load(arg_str)
        args.append(parsed_arg)
        remaining_str = remaining_str[colon_idx+1:]
    return args, from_string(remaining_str)


__version__ = 1.0.3

from .resolvers.default_resolver import DefaultResolver
from .resolvers.resolver import Resolver
from .resolvers.scalar_resolvers.scalar_resolver import ScalarResolver
from .resolvers.collection_resolvers.collection_resolver import CollectionResolver
from .resolvers.post_processors.post_processor import PostProcessor

from .functional.load import from_string, from_file_uri


from .util import apply, accessors_to_string, string_to_accessors
from ..entities.wrappers import KCObject
import yaml

def from_dotlist(dotlist):
    result = dict(root={})
    for entry in dotlist:
        accessor_value_split = entry.split("=")
        if len(accessor_value_split) != 2:
            from..errors import dotlist_entry_requires_equal_sign_error
            raise dotlist_entry_requires_equal_sign_error()
        accessor_string, value = accessor_value_split
        accessors = string_to_accessors(accessor_string)

        # create missing parent objects
        prev_node = result["root"]
        for accessor_idx in range(len(accessors[:-1])):
            cur_accessor = accessors[accessor_idx]
            next_accessor = accessors[accessor_idx + 1]

            # lists can only be created in sequential order
            if isinstance(cur_accessor, int):
                if len(prev_node) != cur_accessor:
                    from ..errors import dotlist_requires_sequential_insert_error
                    raise dotlist_requires_sequential_insert_error()

            # create missing datastructures
            if isinstance(next_accessor, int):
                # create list (if it doesn't exist already)
                if isinstance(cur_accessor, int):
                    prev_node.append([])
                elif cur_accessor not in prev_node:
                    prev_node[cur_accessor] = []
            else:
                # create dict (if it doesn't exist already)
                if isinstance(cur_accessor, int):
                    prev_node.append({})
                elif cur_accessor not in prev_node:
                    prev_node[cur_accessor] = {}

            # progress to next accessor
            prev_node = prev_node[cur_accessor]

        # parse value
        parsed_value = yaml.safe_load(value)

        # insert current value
        last_accessor = accessors[-1]
        if isinstance(last_accessor, int):
            if len(prev_node) != last_accessor:
                from ..errors import dotlist_requires_sequential_insert_error
                raise dotlist_requires_sequential_insert_error()
            prev_node.append(parsed_value)
        else:
            prev_node[last_accessor] = parsed_value

    return result["root"]


def to_dotlist(root_node):
    if isinstance(root_node, KCObject):
        from ..errors import requires_primitive_node_error
        raise requires_primitive_node_error()
    container = dict(accessors=[], result=[])
    apply(root_node, pre_fn=_to_dotlist_pre_fn, post_fn=_to_dotlist_post_fn, container=container)
    return container["result"]


def _to_dotlist_pre_fn(node, parent_accessor, container, **_):
    if isinstance(node, dict) or isinstance(node, list):
        if parent_accessor is not None:
            container["accessors"].append(parent_accessor)
    else:
        if parent_accessor is None:
            cur_node_accessors = []
        else:
            cur_node_accessors = [parent_accessor]
        accessors = container["accessors"] + cur_node_accessors
        if len(accessors) == 0:
            node_str = str(node)
        else:
            accessor_str = accessors_to_string(container["accessors"] + cur_node_accessors)
            node_str = f"{accessor_str}={str(node)}"
        container["result"].append(node_str)


def _to_dotlist_post_fn(node, parent_accessor, container, **_):
    if isinstance(node, dict) or isinstance(node, list):
        if parent_accessor is not None:
            container["accessors"].pop()



class CollectionResolver:
    def preorder_resolve(self, node, root_node, result, trace, root_resolver):
        pass

    def postorder_resolve(self, node, root_node, result, trace, root_resolver):
        pass


from .resolver import Resolver
from .scalar_resolvers.interpolation_resolver import InterpolationResolver
from .scalar_resolvers.eval_resolver import EvalResolver
from .collection_resolvers.template_resolver import TemplateResolver
from .collection_resolvers.missing_value_resolver import MissingValueResolver
from .scalar_resolvers.nested_yaml_resolver import NestedYamlResolver
from .scalar_resolvers.select_resolver import SelectResolver
from .post_processors.remove_vars_post_processor import RemoveVarsPostProcessor
from .post_processors.remove_nones_post_processor import RemoveNonesPostProcessor


class DefaultResolver(Resolver):
    def __init__(self, template_path=None, **templates):
        super().__init__(
            collection_resolvers=[
                TemplateResolver(template_path=template_path, **(templates or {})),
                MissingValueResolver(),
            ],
            default_scalar_resolver=InterpolationResolver(),
            scalar_resolvers=dict(
                eval=EvalResolver(),
                yaml=NestedYamlResolver(template_path=template_path, **(templates or {})),
                select=SelectResolver(),
            ),
            post_processors=[
                RemoveVarsPostProcessor(),
                RemoveNonesPostProcessor(),
            ]
        )


from ..entities.wrappers import KCDict, KCList, KCScalar, KCObject
from ..functional.parse_grammar import parse_grammar
from ..entities.grammar_tree_nodes import RootNode, FixedNode, InterpolatedNode
from copy import deepcopy

class Resolver:
    def __init__(
            self, collection_resolvers=None,
            scalar_resolvers=None, default_scalar_resolver=None,
            post_processors=None):
        self.collection_resolvers = collection_resolvers or []
        self.scalar_resolvers = scalar_resolvers or {}
        if default_scalar_resolver is not None:
            self.scalar_resolvers[None] = default_scalar_resolver
        self.post_processors = post_processors or []


    def resolve(self, node, root_node=None):
        node = deepcopy(node)
        result = {}
        root_node_to_pass = node if root_node is None else root_node
        wrapped_node = KCDict(root=node)
        self._resolve_collection(node, root_node=root_node_to_pass, result=result, trace=[(wrapped_node, "root")])
        processed_result = result["root"]
        # only postprocess from root call (e.g. template resolver also calls resolve but with a root_node parameter)
        if root_node is None:
            post_processed = self.post_process(processed_result)
        else:
            post_processed = processed_result
        return post_processed


    def _resolve_collection(self, node, root_node, result, trace):
        parent, parent_accessor = trace[-1]
        if isinstance(node, KCDict):
            if isinstance(parent_accessor, int):
                result.append({})
            else:
                result[parent_accessor] = {}
            # preorder
            for resolver in self.collection_resolvers:
                resolver.preorder_resolve(node, root_node=root_node, result=result, trace=trace, root_resolver=self)
                # preorder is allowed to change the current node
                node = parent[parent_accessor]

            # traverse
            for accessor, subnode in node.dict.items():
                trace.append((node, accessor))
                self._resolve_collection(subnode, root_node=root_node, result=result[parent_accessor], trace=trace)
                trace.pop()

            # postorder
            for resolver in self.collection_resolvers:
                resolver.postorder_resolve(node, root_node=root_node, result=result, trace=trace, root_resolver=self)
        elif isinstance(node, KCList):
            if isinstance(parent_accessor, int):
                result[parent_accessor].append([])
            else:
                result[parent_accessor] = []
            # preorder
            for resolver in self.collection_resolvers:
                resolver.preorder_resolve(node, root_node=root_node, result=result, trace=trace, root_resolver=self)
                # preorder is allowed to change the current node
                node = parent[parent_accessor]

            # traverse
            for i, subnode in enumerate(node.list):
                trace.append((node, i))
                self._resolve_collection(subnode, root_node=root_node, result=result[parent_accessor], trace=trace)
                trace.pop()

            # postorder
            for resolver in self.collection_resolvers:
                resolver.postorder_resolve(node, root_node=root_node, result=result, trace=trace, root_resolver=self)
        elif isinstance(node, KCScalar):
            # preorder
            for resolver in self.collection_resolvers:
                resolver.preorder_resolve(node, root_node=root_node, result=result, trace=trace, root_resolver=self)
                # preorder is allowed to change the current node
                node = parent[parent_accessor]

            if not isinstance(node.value, str):
                resolve_result = node.value
            else:
                resolve_result = self.resolve_scalar(node.value, root_node=root_node)

            # resolved value might be a KCObject (e.g. when loading a nested yaml)
            if isinstance(resolve_result, KCObject):
                parent[parent_accessor] = resolve_result
                self._resolve_collection(resolve_result, root_node=root_node, result=result, trace=trace)
            else:
                # set value
                parent_accessor = trace[-1][1]
                if isinstance(parent_accessor, int):
                    if len(result) != parent_accessor:
                        raise IndexError
                    result.append(resolve_result)
                else:
                    result[parent_accessor] = resolve_result

            # postorder
            for resolver in self.collection_resolvers:
                resolver.postorder_resolve(node, root_node=root_node, result=result, trace=trace, root_resolver=self)
        else:
            from ..errors import unexpected_type_error
            raise unexpected_type_error([KCDict, KCList, KCScalar], node)

    def resolve_scalar(self, value, root_node):
        grammar_tree = parse_grammar(value)
        resolved_scalar = self._resolve_scalar(grammar_tree, root_node=root_node)
        return resolved_scalar

    def _resolve_scalar(self, grammar_node, root_node):
        if isinstance(grammar_node, RootNode):
            resolve_results = [self._resolve_scalar(child, root_node=root_node) for child in grammar_node.children]
            return self._merge_scalar_resolve_results(resolve_results)
        elif isinstance(grammar_node, FixedNode):
            return grammar_node.value
        elif isinstance(grammar_node, InterpolatedNode):
            # resolve children
            resolve_results = [self._resolve_scalar(child, root_node=root_node) for child in grammar_node.children]
            resolve_result = self._merge_scalar_resolve_results(resolve_results)
            # resolve cur node
            if grammar_node.resolver_key not in self.scalar_resolvers:
                from ..errors import invalid_resolver_key
                raise invalid_resolver_key(grammar_node.resolver_key, list(self.scalar_resolvers.keys()))
            scalar_resolver = self.scalar_resolvers[grammar_node.resolver_key]
            resolved_scalar = scalar_resolver.resolve(resolve_result, root_node=root_node)
            return resolved_scalar
        else:
            from ..errors import unexpected_type_error
            raise unexpected_type_error([RootNode, FixedNode, InterpolatedNode], grammar_node)

    @staticmethod
    def _merge_scalar_resolve_results(resolve_results):
        if len(resolve_results) == 1:
            # keep datatype of single result
            return resolve_results[0]
        else:
            # concat as string
            return "".join(map(str, resolve_results))

    def post_process(self, node):
        wrapped = dict(root=node)
        self._post_process(wrapped, trace=[])
        if "root" not in wrapped:
            from ..errors import empty_result
            raise empty_result()
        return wrapped["root"]

    def _post_process(self, node, trace):
        for post_processor in self.post_processors:
            post_processor.preorder_process(node, trace=trace)
        if isinstance(node, dict):
            for key in node.keys():
                trace.append((node, key))
                self._post_process(node[key], trace=trace)
                trace.pop()
        elif isinstance(node, list):
            for i, item in enumerate(node):
                trace.append((node, i))
                self._post_process(item, trace=trace)
        for post_processor in self.post_processors:
            post_processor.postorder_process(node, trace=trace)


from ..entities.wrappers import KCDict, KCList

def apply(node, pre_fn=None, post_fn=None, parent_node=None, parent_accessor=None, container=None):
    # do something before traversing the node
    if pre_fn is not None:
        pre_fn(node=node, parent_node=parent_node, parent_accessor=parent_accessor, container=container)
        # pre_fn might change the config obj
        if parent_node is not None and parent_accessor is not None:
            node = parent_node[parent_accessor]

    # traverse
    if isinstance(node, KCDict) or isinstance(node, dict):
        for key, value in node.items():
            apply(value, pre_fn=pre_fn, post_fn=post_fn, parent_node=node, parent_accessor=key, container=container)
    elif isinstance(node, KCList) or isinstance(node, list):
        for i, item in enumerate(node):
            apply(item, pre_fn=pre_fn, post_fn=post_fn, parent_node=node, parent_accessor=i, container=container)

    # do something after traversing the node
    if post_fn is not None:
        post_fn(node=node, parent_node=parent_node, parent_accessor=parent_accessor, container=container)


def accessors_to_string(accessors):
    if len(accessors) == 0: return ""

    if isinstance(accessors[0], int):
        result = f"[{accessors[0]}]"
    else:
        result = f"{accessors[0]}"
    for accessor in accessors[1:]:
        if isinstance(accessor, int):
            result += f"[{accessor}]"
        else:
            result += f".{accessor}"
    return result

def string_to_accessors(accessor_str):
    if accessor_str == "": return []

    result = []
    dot_splits = accessor_str.split(".")
    for dot_split in dot_splits:
        bracket_splits = dot_split.split("[")
        # add dictionary accessor
        if bracket_splits[0] != "":
            result.append(bracket_splits[0])

        # add list accessors
        for bracket_split in bracket_splits[1:]:
            # remove closing bracket (']')
            if bracket_split[-1] != "]":
                from ..errors import missing_closing_bracket_error
                raise missing_closing_bracket_error(f"[{bracket_split}")
            list_accessor = bracket_split[:-1]
            result.append(int(list_accessor))
    return result

def trace_to_full_accessor(trace):
    accessors = list(map(lambda tr: tr[1], trace[1:]))
    return accessors_to_string(accessors)

def select(root_node, accessors):
    cur_node = root_node
    for i, accessor in enumerate(accessors):
        try:
            cur_node = cur_node[accessor]
        except:
            from ..errors import invalid_accessor_error
            raise invalid_accessor_error(accessors_to_string(accessors[:i+1]), accessors_to_string(accessors))
    return cur_node

def merge(base, to_merge):
    for key, value in to_merge.items():
        accessors = string_to_accessors(key)
        node = select(base, accessors[:-1])

        # allow appending to a list if the last accessor is 'add' or 'append'
        if isinstance(node, (KCList, list)) and accessors[-1] in ["add", "append"]:
            if not isinstance(value, (KCList, list)):
                from ..errors import unexpected_type_error
                raise unexpected_type_error([KCList, list], value)
            node += value
        else:
            node[accessors[-1]] = value
    return base

def mask_out(dict_, keys_to_mask_out):
    masked_dict = type(dict_)()
    for key, value in dict_.items():
        if key not in keys_to_mask_out:
            masked_dict[key] = value
    return masked_dict

def mask_in(dict_, keys_to_mask_in):
    masked_dict = type(dict_)()
    for key, value in dict_.items():
        if key in keys_to_mask_in:
            masked_dict[key] = value
    return masked_dict


def unexpected_type_error(expected_type_or_types, actual_value):
    if isinstance(expected_type_or_types, list):
        expected_type_str = str(list(map(lambda t: t.__name__, expected_type_or_types)))
        expected_type_str = f"[{expected_type_str}]"
        plural_str = "(s)"
    else:
        expected_type_str = f"'{expected_type_or_types.__name__}'"
        plural_str = ""
    msg = f"expected type{plural_str} {expected_type_str} but got '{type(actual_value).__name__}' ({actual_value})"
    return TypeError(msg)

def index_out_of_range_error(idx, max_idx):
    return IndexError(f"index {idx} is out of range (0<={idx}<{max_idx} evaluates to false)")

def dotlist_entry_requires_equal_sign_error():
    return ValueError("every entry in a dotlist requires a '=' character")

def dotlist_requires_sequential_insert_error():
    msg = "constructing a list from a dotlist requires the indices of the list to start at 0 and be in order"
    return ValueError(msg)

def requires_primitive_node_error():
    return TypeError("requires primitive node")

def empty_resolver_key_error(value_in_brace):
    return ValueError(f"empty resolver key in '${{{value_in_brace}}}'")

def empty_resolver_value_error(value_in_brace):
    return ValueError(f"empty resolver value in '${{{value_in_brace}}}'")

def missing_closing_brace_error(value):
    return ValueError(f"missing '}}' in '{value}'")

def missing_closing_bracket_error(value):
    return ValueError(f"missing ']' in '{value}'")

class MissingValueError(Exception):
    pass

def missing_value_error(full_accessor):
    return MissingValueError(full_accessor)

class InvalidAccessorError(Exception):
    pass

def invalid_accessor_error(accessor_until_invalid, full_accessor):
    return InvalidAccessorError(f"invalid accessor '{accessor_until_invalid}' within '{full_accessor}'")

def empty_parameter_error(args_and_value_str):
    return ValueError(f"empty parameter in '{args_and_value_str}'")

def missing_parameter_error(args_and_value_str, n_args):
    return ValueError(f"missing parameter in '{args_and_value_str}', expected {n_args} parameters")

def invalid_resolver_key(resolver_key, valid_keys):
    # don't use key error here as it escapes the whole string and the ' characters
    return ValueError(f"invalid resolver key '{resolver_key}' valid keys are: {valid_keys}")

def empty_result():
    return MissingValueError("empty result")


import yaml
from .convert import from_primitive

def from_file_uri(file_uri):
    with open(file_uri) as f:
        return from_primitive(yaml.safe_load(f))

def from_string(yaml_string):
    if not isinstance(yaml_string, str):
        from ..errors import unexpected_type_error
        raise unexpected_type_error(str, yaml_string)
    return from_primitive(yaml.safe_load(yaml_string))


class GrammarTreeNode:
    def __repr__(self):
        raise NotImplementedError

    def __str__(self):
        return repr(self)

class RootNode(GrammarTreeNode):
    def __init__(self):
        super().__init__()
        self.children = []

    def __repr__(self):
        return "".join(map(str, self.children))

class FixedNode(GrammarTreeNode):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def __repr__(self):
        return self.value

class InterpolatedNode(GrammarTreeNode):
    def __init__(self, resolver_key, *args):
        super().__init__()
        self.children = []
        self.resolver_key = resolver_key
        self.args = args

    def __repr__(self):
        children_repr = str(list(map(str, self.children)))
        resolver_repr = f"{self.resolver_key}:" if self.resolver_key else ""
        return f"${{{resolver_repr}{children_repr}}}"


import unittest
import os
import kappaconfig as kc
import yaml

class TestComplexYamls(unittest.TestCase):
    @staticmethod
    def get_file_uris(root_path):
        file_or_dir_list = os.listdir(root_path)
        file_uris = []
        for file_or_dir in file_or_dir_list:
            file_or_dir_uri = f"{root_path}/{file_or_dir}"
            if file_or_dir_uri.endswith(".result.yaml"): continue

            if os.path.isdir(file_or_dir_uri):
                file_uris += TestComplexYamls.get_file_uris(file_or_dir_uri)
            else:
                file_uris += [file_or_dir_uri]
        return file_uris


    def test_testcase_exists_for_every_yaml_file(self):
        # there probably are libraries that allow this implicitly (e.g. ddt should be able to do this)
        # but this makes running/debugging individual tests a hassle, so just check it manually
        root_path = "tests_integration/complex_yamls"
        file_uris = self.get_file_uris(root_path)
        file_uris = list(map(lambda fu: fu[len(root_path) + 1:], file_uris))
        with open(__file__) as f:
            yaml_files_test_content = f.read()
        # basic check if filename is contained
        for file_uri in file_uris:
            self.assertTrue(file_uri in yaml_files_test_content, f"no testcase for {file_uri}")

    def resolve_yaml(self, src_file_name):
        kc_obj = kc.from_file_uri(f"tests_integration/complex_yamls/{src_file_name}")
        resolver = kc.DefaultResolver(template_path="tests_integration/templates")
        expected_file_name = src_file_name.replace(".yaml", "") + ".result.yaml"
        with open(f"tests_integration/complex_yamls/{expected_file_name}") as f:
            expected = yaml.safe_load(f)
        self.assertEqual(expected, resolver.resolve(kc_obj))

    def test_data_cifar(self):
        self.resolve_yaml("data/cifar.yaml")

    def test_data_mvtec(self):
        self.resolve_yaml("data/mvtec.yaml")

    def test_loggers_default_epochs(self):
        self.resolve_yaml("loggers/default_epochs.yaml")

    def test_loggers_default_updates(self):
        self.resolve_yaml("loggers/default_updates.yaml")

    def test_loggers_discriminator_epochs(self):
        self.resolve_yaml("loggers/discriminator_epochs.yaml")

    def test_models_mae(self):
        self.resolve_yaml("models/mae.yaml")

    def test_models_vit(self):
        self.resolve_yaml("models/vit.yaml")

    def test_optim_mae_pretrain(self):
        self.resolve_yaml("optims/mae_pretrain.yaml")

    def test_optim_scaled_lr(self):
        self.resolve_yaml("optims/scaled_lr.yaml")

    def test_schedule_warmup_cosine(self):
        self.resolve_yaml("schedules/warmup_cosine.yaml")



from .post_processor import PostProcessor

class RemoveVarsPostProcessor(PostProcessor):
    def postorder_process(self, node, **_):
        if isinstance(node, dict) and "vars" in node:
            del node["vars"]


from .collection_resolver import CollectionResolver
from ...entities.wrappers import KCDict, KCScalar
from ...functional.util import mask_in, mask_out, merge
from ...functional.convert import from_primitive
from ..resolver import Resolver
from ..scalar_resolvers.nested_yaml_resolver import NestedYamlResolver
from ..scalar_resolvers.interpolation_resolver import InterpolationResolver
from ..scalar_resolvers.select_resolver import SelectResolver
from ..scalar_resolvers.eval_resolver import EvalResolver

class TemplateResolver(CollectionResolver):
    def __init__(self, template_path=None, **templates):
        super().__init__()
        # nested yaml templates are resolved after loading/merging with template parameters
        # this should not resolve templates or throw missing value errors (these have to happen after the
        # template params are passed to the template)
        self.nested_yaml_resolver = Resolver(
            default_scalar_resolver=InterpolationResolver(),
            scalar_resolvers=dict(
                eval=EvalResolver(),
                yaml=NestedYamlResolver(template_path=template_path, **templates),
                select=SelectResolver(),
            ),
        )

    def preorder_resolve(self, node, root_node, result, trace, root_resolver):
        if isinstance(node, KCDict):
            while True:
                if "template" not in node:
                    break
                template = node["template"]

                if isinstance(template, KCDict):
                    # template is a dictionary template (nothing needs to be resolved)
                    # dict_template:
                    #   template:
                    #     key: value
                    resolved_template = template
                    node_without_template_params = node
                elif isinstance(template, KCScalar):
                    # Template is a string template. String templates need to be self contained but can expect
                    # parameters to be "passed" to them.
                    # string_template:
                    #   template: ${yaml:some_template.yaml}
                    # string_var_template:
                    #   template: ${yaml:some_template.yaml}
                    #   template.vars:
                    #     param: 5
                    resolved_scalar = self.nested_yaml_resolver.resolve_scalar(template, root_node=root_node)
                    # merge template parameters into template
                    template_params_keys = list(filter(lambda key: key.startswith("template."), node.keys()))
                    template_params = mask_in(node, template_params_keys)
                    # resolve template params (for recursive parameter passing)
                    resolved_template_params = {}
                    for k, v in template_params.items():
                        new_key = k.replace("template.", "")
                        if isinstance(v, KCScalar):
                            v = root_resolver.resolve(v, root_node=root_node)
                            v = from_primitive(v)
                        resolved_template_params[new_key] = v

                    kc_resolved_scalar = from_primitive(resolved_scalar)
                    merged_template = merge(kc_resolved_scalar, resolved_template_params)

                    # resolve template with template root as root (for resolving parameterized templates)
                    resolved_template_primitive = root_resolver.resolve(merged_template)
                    # remove "vars" field in template
                    resolved_template_primitive_no_vars = mask_out(resolved_template_primitive, ["vars"])
                    resolved_template = from_primitive(resolved_template_primitive_no_vars)

                    # remove template params from node
                    node_without_template_params = mask_out(node, template_params_keys)
                else:
                    # KCList not implemented yet
                    raise NotImplementedError


                masked = mask_out(node_without_template_params, ["template"])
                merged = merge(resolved_template, masked)
                parent, accessor = trace[-1]
                parent[accessor] = merged
                node = merged







from .scalar_resolver import ScalarResolver
from ...functional.util import string_to_accessors, select
from ...functional.parse_grammar import parse_resolver_args_and_value

class SelectResolver(ScalarResolver):
    """
    allows selecting fields within a nested yaml import
    e.g. model: ${select:base:${yaml:models/vit}}
    """

    def resolve(self, args_and_value, **_):
        args, value = parse_resolver_args_and_value(args_and_value, n_args=1)
        accessors = string_to_accessors(args[0])
        node = select(root_node=value, accessors=accessors)
        return node


from .post_processor import PostProcessor

class RemoveNonesPostProcessor(PostProcessor):
    """
    removes nodes that are set to None
    e.g.
    save_logger:
      every_n_epochs: 1
      every_n_updates: None
    resolves to:
    save_logger:
      every_n_epochs: 1
    """
    def preorder_process(self, node, **_):
        if isinstance(node, dict):
            keys = list(node.keys())
            for key in keys:
                if node[key] is None:
                    del node[key]
        elif isinstance(node, list):
            i = 0
            while i < len(node):
                if node[i] is None:
                    del node[i]
                else:
                    i += 1